import abc
import asyncio
import json
import os
from json import JSONDecodeError
from typing import Dict, NoReturn, Set

from loguru import logger
from pydantic import ValidationError
from starbot_executor import executor

from .. import config
from ..core.event import EventType, DataSourceEvent
from ..core.model import Up
from ..exception.DataSourceException import DataSourceException


class DataSource(metaclass=abc.ABCMeta):
    """
    推送配置数据源基类
    """

    def __init__(self):
        self.ups: Set[Up] = set()
        self.uids: Set[int] = set()
        self.__up_map: Dict[int, Up] = {}

    def __getitem__(self, key):
        return self.__up_map[key]

    @abc.abstractmethod
    async def load(self) -> NoReturn:
        """
        读取配置，基类空实现
        """
        pass

    def add(self, up: Up) -> NoReturn:
        """
        动态添加主播

        Args:
            up: 主播实例
        """
        if up.uid in self.uids:
            raise DataSourceException(f"数据源中不可含有重复的主播 (UID: {up.uid})")

        self.ups.add(up)
        self.uids.add(up.uid)
        self.__up_map[up.uid] = up

        executor.dispatch(up, EventType.DataSourceEvent, DataSourceEvent.DataSourceAdded)

    def remove(self, uid: int) -> NoReturn:
        """
        动态移除主播

        Args:
            uid: 主播 UID
        """
        if uid not in self.uids:
            raise DataSourceException(f"主播 (UID: {uid}) 不存在于数据源中")

        up = self.__up_map.get(uid)
        self.ups.remove(up)
        self.uids.remove(uid)
        self.__up_map.pop(uid)

        executor.dispatch(up, EventType.DataSourceEvent, DataSourceEvent.DataSourceRemoved)

    def update(self, up: Up) -> NoReturn:
        """
        动态更新主播

        Args:
            up: 主播实例
        """
        if up.uid not in self.uids:
            raise DataSourceException(f"主播 (UID: {up.uid}) 不存在于数据源中")

        old = self.__up_map.get(up.uid)
        self.ups.remove(old)
        self.uids.remove(old.uid)
        self.__up_map.pop(old.uid)
        self.ups.add(up)
        self.uids.add(up.uid)
        self.__up_map[up.uid] = up

        executor.dispatch(up, EventType.DataSourceEvent, DataSourceEvent.DataSourceUpdated)


class JsonDataSource(DataSource):
    """
    使用 JSON 初始化的推送配置数据源
    """
    def __init__(self):
        super().__init__()
        self.__config = None

        self.__json_file = config.get("datasource.json_datasource.file_path", str, "推送配置.json")
        self.__auto_reload = config.get("datasource.json_datasource.auto_reload", bool, True)
        self.__auto_reload_interval = config.get("datasource.json_datasource.auto_reload_interval", int, 5)

    async def load(self) -> NoReturn:
        """
        从 JSON 字符串中初始化配置
        """
        if self.ups:
            return

        logger.info("已选用 JSON 作为 Bot 数据源")
        logger.info("开始从 JSON 中初始化 Bot 配置")

        try:
            modify_time = os.path.getmtime(self.__json_file)
            with open(self.__json_file, encoding="utf-8") as file:
                content = file.read()
        except FileNotFoundError:
            raise DataSourceException("JSON 文件不存在, 请检查文件路径是否正确")
        except UnicodeDecodeError:
            raise DataSourceException("JSON 文件编码不正确, 请将其转换为 UTF-8 格式编码后重试")
        except Exception as ex:
            raise DataSourceException(f"读取 JSON 文件异常 {ex}")

        try:
            self.__config = json.loads(content)
        except Exception:
            raise DataSourceException("JSON 文件内容格式不正确")

        if isinstance(self.__config, dict):
            self.__config = [self.__config]

        for up in self.__config:
            try:
                self.add(Up(**up))
            except ValidationError as ex:
                raise DataSourceException(f"提供的配置字典中缺少必须的 {ex.errors()[0].get('loc')[-1]} 参数")

        logger.success(f"成功从 JSON 中导入了 {len(self.ups)} 个 UP 主")

        if self.__auto_reload:
            executor.create_task(self.__auto_reload_task(modify_time))

    async def __auto_reload_task(self, modify_time: float) -> NoReturn:
        """
        JSON 文件内容发生变化时自动重载配置

        Args:
            modify_time: 配置文件修改时间
        """
        while True:
            await asyncio.sleep(self.__auto_reload_interval)
            try:
                new_modify_time = os.path.getmtime(self.__json_file)

                if new_modify_time == modify_time:
                    continue

                modify_time = new_modify_time

                logger.info(f"数据源配置已更新, 开始重载配置")

                with open(self.__json_file, encoding="utf-8") as file:
                    content = file.read()

                conf = json.loads(content)
                if isinstance(conf, dict):
                    conf = [conf]

                new_ups = []
                for up in conf:
                    new_ups.append(Up(**up))

                new_up_set = set(new_ups)
                if len(new_up_set) != len(new_ups):
                    raise DataSourceException("数据源中不可含有重复的主播")

                self.__config = conf

                added_ups = new_up_set - self.ups
                removed_ups = self.ups - new_up_set

                tip = [f'{act}了 {len(ups)} 个主播' for act, ups in [('新增', added_ups), ('移除', removed_ups)] if ups]
                if tip:
                    logger.info(f"检测到 {', '.join(tip)}")
                else:
                    logger.info("未检测到新增或移除的主播, 开始重载已存在主播的配置")

                for up in new_up_set.union(self.ups):
                    if up in added_ups:
                        self.add(up)
                    elif up in removed_ups:
                        self.remove(up.uid)
                    else:
                        self.update(up)

                logger.success("数据源配置重载成功")
            except FileNotFoundError:
                logger.error("数据源配置 JSON 文件不存在")
            except UnicodeDecodeError:
                logger.error("数据源配置 JSON 文件编码不正确, 请将其转换为 UTF-8 格式编码")
            except JSONDecodeError:
                logger.error("数据源配置 JSON 文件内容格式不正确")
            except DataSourceException as ex:
                logger.error(ex.msg)
            except Exception as ex:
                logger.error(f"数据源自动重载任务异常 {ex}")
            finally:
                continue



import json
from json import JSONDecodeError
from typing import Dict, Any, Type, Optional

from loguru import logger

from ..exception.DataSourceException import DataSourceException


class Config:
    """
    全局配置类
    """

    __config: Dict[str, Any]

    def __init__(self):
        try:
            with open("config.json", encoding="utf-8") as file:
                self.__config = json.loads(file.read())
        except Exception as ex:
            if isinstance(ex, FileNotFoundError):
                logger.error("配置文件 config.json 不存在")
            elif isinstance(ex, UnicodeDecodeError):
                logger.error("配置文件 config.json 编码不正确, 请将其转换为 UTF-8 格式编码")
            elif isinstance(ex, JSONDecodeError):
                logger.error("配置文件 config.json 内容格式不正确")
            else:
                logger.error(f"读取配置文件 config.json 异常 {ex}")
            raise DataSourceException("读取配置文件 config.json 异常")

    def get(self, attribute: str, _type: Type, default: Optional[Any] = None) -> Optional[Any]:
        """
        获取配置项的值

        Args:
            attribute: 配置项，多个配置项之间可使用.分隔
            _type: 预期类型，若类型不匹配会产生警告，并尝试进行自动转换
            default: 默认值，若要获取的配置项不存在或类型转换失败，返回默认值。默认：None

        Returns:
            配置项的值，若要获取的配置项不存在或类型转换失败，返回默认值
        """
        keys = attribute.split(".")
        conf = self.__config

        for key in keys:
            if not isinstance(conf, dict):
                logger.warning(f"配置项 {attribute} 不存在, 已使用默认值: {default}, 请补全配置文件 config.json")
                return default

            if key not in conf:
                logger.warning(f"配置项 {attribute} 不存在, 已使用默认值: {default}, 请补全配置文件 config.json")
                return default

            conf = conf[key]

        if isinstance(conf, _type):
            return conf

        logger.warning(
            f"配置项 {attribute} 数据类型不正确, 预期类型: {_type}, 实际类型: {type(conf)}, 请检查配置文件 config.json"
        )

        try:
            auto_cast_value = _type(conf)
            logger.success(f"配置项 {attribute} 数据类型自动转换成功")
            return auto_cast_value
        except Exception:
            logger.warning(f"配置项 {attribute} 数据类型自动转换失败, 已使用默认值: {default}, 请检查配置文件 config.json")
            return default


config: Config = Config()



from enum import Enum


class EventType:
    """
    事件类型枚举，基于本项目开发时，可通过继承此类的方式扩展事件类型
    """
    DataSourceEvent = "DataSourceEvent"
    """数据源事件"""


class DataSourceEvent(Enum):
    """
    数据源事件类型枚举
    """
    DataSourceAdded = "DataSourceAdded"
    """数据源主播新增"""

    DataSourceRemoved = "DataSourceRemoved"
    """数据源主播移除"""

    DataSourceUpdated = "DataSourceUpdated"
    """数据源主播更新"""



from typing import Optional, List, Union

from pydantic import BaseModel


class LiveOn(BaseModel):
    """
    开播推送配置
    可使用构造方法手动传入所需的各项配置
    或使用 LiveOn.default() 获取功能全部开启的默认配置
    """

    enabled: bool = False
    """是否启用开播推送。默认：False"""

    message: str = ""
    """
    开播推送内容模板。
    专用占位符: {uname} 主播昵称，{title} 直播间标题，{url} 直播间链接，{cover} 直播间封面图。
    通用占位符: {next} 消息分条，{atall} @全体成员，{at=114514} @指定成员，
              {urlpic=链接} 网络图片，{pathpic=路径} 本地图片，{base64pic=base64字符串} base64图片。
    默认：""
    """

    @classmethod
    def default(cls):
        """
        获取功能全部开启的默认 LiveOn 实例
        默认配置：启用开播推送，推送内容模板为 "{uname} 正在直播 {title}\n{url}{next}{cover}"
        """
        return LiveOn(enabled=True, message="{uname} 正在直播 {title}\n{url}{next}{cover}")


class LiveOff(BaseModel):
    """
    下播推送配置
    可使用构造方法手动传入所需的各项配置
    或使用 LiveOff.default() 获取功能全部开启的默认配置
    """

    enabled: bool = False
    """是否启用下播推送。默认：False"""

    message: str = ""
    """
    下播推送内容模板。
    专用占位符: {uname} 主播昵称。
    通用占位符: {next} 消息分条，{atall} @全体成员，{at=114514} @指定成员，
              {urlpic=链接} 网络图片，{pathpic=路径} 本地图片，{base64pic=base64字符串} base64图片。
    默认：""
    """

    @classmethod
    def default(cls):
        """
        获取功能全部开启的默认 LiveOff 实例
        默认配置：启用下播推送，推送内容模板为 "{uname} 直播结束了"
        """
        return LiveOff(enabled=True, message="{uname} 直播结束了")


class LiveReport(BaseModel):
    """
    直播报告配置，直播报告会在下播推送后发出，下播推送是否开启不会影响直播报告的推送
    可使用构造方法手动传入所需的各项配置
    或使用 LiveReport.default() 获取功能全部开启的默认配置
    """

    enabled: bool = False
    """是否启用直播报告。默认：False"""

    logo: Optional[str] = None
    """主播立绘的路径，会绘制在直播报告右上角合适位置。默认：None"""

    logo_base64: Optional[str] = None
    """主播立绘的 Base64 字符串，会绘制在直播报告右上角合适位置，立绘路径不为空时优先使用路径。默认：None"""

    time: bool = False
    """是否展示本场直播直播时间段和直播时长。默认：False"""

    fans_change: bool = False
    """是否展示本场直播粉丝变动。默认：False"""

    fans_medal_change: bool = False
    """是否展示本场直播粉丝团（粉丝勋章数）变动。默认：False"""

    guard_change: bool = False
    """是否展示本场直播大航海变动。默认：False"""

    danmu: bool = False
    """是否展示本场直播收到弹幕数、发送弹幕人数。默认：False"""

    box: bool = False
    """是否展示本场直播收到盲盒数、送出盲盒人数、盲盒盈亏。默认：False"""

    gift: bool = False
    """是否展示本场直播礼物收益、送礼物人数。默认：False"""

    sc: bool = False
    """是否展示本场直播 SC（醒目留言）收益、发送 SC（醒目留言）人数。默认：False"""

    guard: bool = False
    """是否展示本场直播开通大航海数。默认：False"""

    danmu_ranking: int = 0
    """展示本场直播弹幕排行榜的前多少名，0 为不展示。默认：0"""

    box_ranking: int = 0
    """展示本场直播盲盒数量排行榜的前多少名，0 为不展示。默认：0"""

    box_profit_ranking: int = 0
    """展示本场直播盲盒盈亏排行榜的前多少名，0 为不展示。默认：0"""

    gift_ranking: int = 0
    """展示本场直播礼物排行榜的前多少名，0 为不展示。默认：0"""

    sc_ranking: int = 0
    """展示本场直播 SC（醒目留言）排行榜的前多少名，0 为不展示。默认：0"""

    guard_list: bool = False
    """是否展示本场直播开通大航海观众列表。默认：False"""

    box_profit_diagram: bool = False
    """是否展示本场直播的盲盒盈亏曲线图。默认：False"""

    danmu_diagram: bool = False
    """是否展示本场直播的弹幕互动曲线图。默认：False"""

    box_diagram: bool = False
    """是否展示本场直播的盲盒互动曲线图。默认：False"""

    gift_diagram: bool = False
    """是否展示本场直播的礼物互动曲线图。默认：False"""

    sc_diagram: bool = False
    """是否展示本场直播的 SC（醒目留言）互动曲线图。默认：False"""

    guard_diagram: bool = False
    """是否展示本场直播的开通大航海互动曲线图。默认：False"""

    danmu_cloud: bool = False
    """是否生成本场直播弹幕词云。默认：False。默认：False"""

    @classmethod
    def default(cls):
        """
        获取功能全部开启的默认 LiveReport 实例
        默认配置：启用直播报告，无主播立绘，展示直播时间段和直播时长，展示粉丝变动，展示粉丝团（粉丝勋章数）变动，展示大航海变动
                展示收到弹幕数、发送弹幕人数，展示收到盲盒数、送出盲盒人数、盲盒盈亏，展示礼物收益、送礼物人数
                展示 SC（醒目留言）收益、发送 SC（醒目留言）人数，展示开通大航海数
                展示弹幕排行榜前 3 名，展示盲盒数量排行榜前 3 名，展示盲盒盈亏排行榜前 3 名，展示礼物排行榜前 3 名
                展示 SC（醒目留言）排行榜前 3 名，展示开通大航海观众列表
                展示盲盒盈亏曲线图，展示弹幕互动曲线图，展示盲盒互动曲线图，展示礼物互动曲线图，
                展示 SC（醒目留言）互动曲线图，展示开通大航海互动曲线图，
                生成弹幕词云
        """
        return LiveReport(enabled=True, logo=None, logo_base64=None,
                          time=True, fans_change=True, fans_medal_change=True, guard_change=True,
                          danmu=True, box=True, gift=True, sc=True, guard=True,
                          danmu_ranking=3, box_ranking=3, box_profit_ranking=3, gift_ranking=3, sc_ranking=3,
                          guard_list=True, box_profit_diagram=True,
                          danmu_diagram=True, box_diagram=True, gift_diagram=True, sc_diagram=True, guard_diagram=True,
                          danmu_cloud=True)


class DynamicUpdate(BaseModel):
    """
    动态推送配置
    可使用构造方法手动传入所需的各项配置
    或使用 DynamicUpdate.default() 获取功能全部开启的默认配置
    """

    enabled: bool = False
    """是否启用动态推送。默认：False"""

    message: str = ""
    """
    动态推送内容模板。
    专用占位符: {uname} 主播昵称，{action} 动态操作类型（发表了新动态，转发了新动态，投稿了新视频...），
              {url} 动态链接（若为发表视频、专栏等则为视频、专栏等对应的链接），{picture} 动态图片。
    通用占位符: {next} 消息分条，{atall} @全体成员，{at=114514} @指定成员，
              {urlpic=链接} 网络图片，{pathpic=路径} 本地图片，{base64pic=base64字符串} base64图片。
    默认：""
    """

    @classmethod
    def default(cls):
        """
        获取功能全部开启的默认 DynamicUpdate 实例
        默认配置：启用动态推送，推送内容模板为 "{uname} {action}\n{url}{next}{picture}"
        """
        return DynamicUpdate(enabled=True, message="{uname} {action}\n{url}{next}{picture}")


class Platform(BaseModel):
    """
    推送平台类
    """
    name: str
    """推送平台唯一标识符，请使用 平台名称/自定义名称(建议使用推送平台实现所在的代码仓库名) 的格式，并注意唯一性，例：QQ/StarBot"""

    account: Union[int, str]
    """机器人账号"""

    def __eq__(self, other):
        if isinstance(other, Platform):
            return self.name == other.name and self.account == other.account
        return False

    def __hash__(self):
        return hash(self.name) ^ hash(self.account)


class PushTarget(BaseModel):
    """
    推送目标类
    """

    id: Union[int, str]
    """推送目标标识符，一般为账号或群号"""

    platform: Platform
    """推送平台"""

    live_on: LiveOn = LiveOn()
    """开播推送配置。默认：LiveOn()"""

    live_off: LiveOff = LiveOff()
    """下播推送配置。默认：LiveOff()"""

    live_report: LiveReport = LiveReport()
    """直播报告配置。默认：LiveReport()"""

    dynamic_update: DynamicUpdate = DynamicUpdate()
    """动态推送配置。默认：DynamicUpdate()"""

    def __eq__(self, other):
        if isinstance(other, PushTarget):
            return self.id == other.id and self.platform == other.platform
        return False

    def __hash__(self):
        return hash(self.id) ^ hash(self.platform)


class Up(BaseModel):
    """
    主播类
    """

    uid: int
    """主播 UID"""

    targets: List[PushTarget]
    """主播所需推送的推送目标"""

    def __eq__(self, other):
        if isinstance(other, Up):
            return self.uid == other.uid
        return False

    def __hash__(self):
        return hash(self.uid)

    def is_need_connect(self) -> bool:
        """
        根据开播、下播、直播报告开关，判断是否需要连接直播间

        Returns:
            是否需要连接直播间
        """
        return any([
            any(map(lambda conf: conf.enabled, map(lambda group: group.live_on, self.targets))),
            any(map(lambda conf: conf.enabled, map(lambda group: group.live_off, self.targets))),
            any(map(lambda conf: conf.enabled, map(lambda group: group.live_report, self.targets)))
        ])
